from ..utils import *


class BaseAgent:
    def __init__(self, config):
        self.config = config
        self.task_ind = 0

    def close(self):
        close_obj(self.task)

    def eval_step(self, state, reward, done, info, ep=None):
        raise NotImplementedError

    def eval_episode(self, ep):
        task = self.config.eval_task
        state = task.reset()
        reward = 0
        done = True
        info = None
        discounted_ret = 0
        discounted_compound_ret = 0
        discounted_option_terminations = 0
        discounted_interests = 0
        discount_prod = 1
        while True:
            action, option_terminate, interests = self.eval_step(state, reward, done, info, ep)
            discounted_compound_ret += discount_prod * (reward + interests * self.config.beta_reg) * self.config.discount
            state, reward, done, info = task.step(action)
            ret = info['episodic_return']
            len = info['episodic_length']
            discounted_ret += discount_prod * reward
            discounted_option_terminations += discount_prod * (-option_terminate)
            discounted_interests += discount_prod * (-interests)
            discount_prod *= self.config.discount
            if done:
                break
        return ret, len, discounted_ret, discounted_compound_ret, discounted_option_terminations, discounted_interests

    def eval_episodes(self):
        episodic_returns = []
        episodic_lengths = []
        episodic_discounted_returns = []
        episodic_discounted_compound_returns = []
        episodic_discounted_terminations = []
        episodic_discounted_interests = []
        for ep in range(self.config.eval_episodes):
            ret, len, discounted_ret, discounted_compound_ret, discounted_option_terminations, discounted_interests = self.eval_episode(ep)
            episodic_returns.append(ret)
            episodic_lengths.append(len)
            episodic_discounted_returns.append(discounted_ret)
            episodic_discounted_compound_returns.append(discounted_compound_ret)
            episodic_discounted_terminations.append(discounted_option_terminations)
            episodic_discounted_interests.append(discounted_interests)
        return episodic_returns, episodic_lengths, episodic_discounted_returns, episodic_discounted_compound_returns, episodic_discounted_terminations, episodic_discounted_interests