from ..component import *
from .BaseAgent import *


def sigmoid(x):
	return 1 / (1 + np.exp(-x))


def softmax(x):
	e_x = np.exp(x - np.max(x, axis=-1, keepdims=True))
	return e_x / e_x.sum(axis=-1, keepdims=True)


class FPOC(BaseAgent):
	def __init__(self, config):
		BaseAgent.__init__(self, config)
		self.config = config
		self.task = config.task_fn(config.seed)
		self.eval_task = config.eval_task
		self.num_tasks = self.task.env.num_tasks
		self.num_states = self.task.env.num_states
		self.num_actions = self.task.env.num_actions
		self.num_adj_options = self.config.num_adj_options
		if self.config.action_as_option:
			self.num_options = self.num_adj_options + self.num_actions
		else:
			self.num_options = self.num_adj_options
		self.qs = np.zeros((self.num_tasks, self.num_states, self.num_options))
		self.pi_logit = np.zeros((self.num_states, self.num_options, self.num_actions))
		self.beta_logit = np.zeros((self.num_states, self.num_options))
		self.interest_logit = np.zeros((self.num_states, self.num_options))
		if self.config.use_interest is False:
			self.interest_logit[:, :] = 20
		if self.config.action_as_option:
			for a in range(self.num_actions):
				self.interest_logit[:, self.num_adj_options + a] = 20
				self.pi_logit[:, self.num_adj_options + a, a] = 20
				self.beta_logit[:, self.num_adj_options + a] = 20
		self.total_steps = 0
		self.compound_return = 0
		self.state = self.task.reset()
		self.s_is_initial_state = True
		self.prev_option = 0
		self.prev_option_test = None
		self.eps_option_list = []
		if self.config.Option_Set_in_VI == "given_hallway_options_and_actions":
			self.vi_num_options = 2 + self.num_actions
		elif self.config.Option_Set_in_VI == "learned_options_and_actions":
			self.vi_num_options = self.num_options
		elif self.config.Option_Set_in_VI == "actions":
			self.vi_num_options = self.num_actions
		else:
			raise NotImplementedError
		self.train_tasks = self.task.env.goals
		self.test_tasks = [2,7,20,24,36,40,43,54,59,70,73,77,89,93,96,101]
		self.new_task_q_train = np.random.randn(len(self.train_tasks), self.task.env.num_states,
												self.vi_num_options) + np.ones(
			(len(self.train_tasks), self.task.env.num_states, self.vi_num_options)) * -100
		self.new_task_q_test = np.random.randn(len(self.test_tasks), self.task.env.num_states,
											   self.vi_num_options) + np.ones(
			(len(self.test_tasks), self.task.env.num_states, self.vi_num_options)) * -100
		self.option_dynamic_model_train = np.zeros((len(self.train_tasks), self.task.env.num_states, self.vi_num_options, self.task.env.num_states))
		self.option_dynamic_model_test = np.zeros((len(self.test_tasks), self.task.env.num_states, self.vi_num_options, self.task.env.num_states))
		self.option_reward_model_train = np.zeros((len(self.train_tasks), self.task.env.num_states, self.vi_num_options))
		self.option_reward_model_test = np.zeros((len(self.test_tasks), self.task.env.num_states, self.vi_num_options))
		self.best_option1_pi_vec = np.array([1,1,1,1,1,3,1,1,1,1,1,1,1,1,1,3,1,1,1,1,3,3,3,3,3,3,3,1,1,1,1,0,0,0,0,0,3,1,1,1,1,0,0,0,0,0,3,1,1,1,1,0,3,3,1,2,2,3,0,2,2,2,1,0,0,0,0,2,1,1,1,1,1,0,0,0,0,2,1,1,1,1,1,0,0,0,0,2,2,2,2,2,2,2,0,0,0,0,2,0,0,0,0,0])
		self.best_option2_pi_vec = np.array([3,1,2,2,2,1,2,2,2,2,3,1,2,2,2,1,2,2,2,2,3,1,2,2,2,2,2,2,2,2,2,3,1,2,2,2,0,2,2,2,2,3,1,2,2,2,0,2,2,2,2,1,0,0,0,0,0,1,1,1,1,1,0,3,3,3,3,1,3,3,0,2,2,3,3,3,3,1,3,0,0,0,0,3,3,3,3,3,3,3,0,0,0,0,3,3,3,3,0,3,0,0,0,0])
		self.best_option1_beta_vec = np.zeros(104)
		self.best_option1_beta_vec[26] = 1
		self.best_option1_beta_vec[42] = 1
		self.best_option1_beta_vec[70] = 1
		self.best_option1_beta_vec[87] = 1
		self.best_option2_beta_vec = np.zeros(104)
		self.best_option2_beta_vec[24] = 1
		self.best_option2_beta_vec[58] = 1
		self.best_option2_beta_vec[54] = 1
		self.best_option2_beta_vec[89] = 1
		self.current_interests, self.current_initiation_set = self.choose_initiation_set(self.state)

	def choose_option_without_interest(self, task_id, state, prev_option, is_intial_state, epsilon, num_options, qs):
		"""
		This function samples an option according to the task's meta-policy.
		"""
		if epsilon > np.random.rand(1):
			candidate_option = np.random.choice(num_options, 1, p=np.ones(num_options) / num_options)[0]
		else:
			candidate_option = np.random.choice(np.flatnonzero(qs[task_id, state] == qs[task_id, state].max()))

		if is_intial_state is True:
			option = candidate_option
			return option, True
		else:
			beta = sigmoid(self.beta_logit[state, prev_option])
			prev_option_term = np.random.choice(2, 1, p=[1 - beta, beta])[0]

			if prev_option_term:
				option = candidate_option
			else:
				option = prev_option
			return option, prev_option_term

	def choose_initiation_set(self, state):
		if self.config.use_interest is False:
			interests = np.ones(self.num_options)
			initiation_set = np.ones(self.num_options)
			return interests, initiation_set
		interests = sigmoid(self.interest_logit[state])
		initiation_set = np.array(np.random.binomial(1, interests), dtype=bool)
		return interests, initiation_set

	def compute_k_b_probs(self, k_initiation_sets, task_id, state, epsilon):
		k = k_initiation_sets.shape[0]
		num_options_in_set = np.sum(k_initiation_sets, axis=1)  # primitive actions must be in the set
		b_probs = np.zeros((k, self.num_options))
		b_probs += epsilon / np.expand_dims(num_options_in_set, 1) * k_initiation_sets
		tmp = self.qs[task_id, state] - 1e8 * (1 - k_initiation_sets)
		best_options = np.argmax(tmp, axis=1)
		b_probs[np.arange(k), best_options] += (1 - epsilon)
		return b_probs

	def choose_option(self, task_id, state, prev_option, is_intial_state, initiation_set, epsilon, num_options, qs):
		if self.config.use_interest == False:
			option, prev_option_term = self.choose_option_without_interest(task_id, state, prev_option, is_intial_state, epsilon, num_options, qs)
			return option, prev_option_term

		num_options_in_set = sum(initiation_set)  # primitive actions must be in the set
		b_probs = np.zeros(num_options)
		b_probs += epsilon / num_options_in_set * np.array(initiation_set)
		tmp = qs[task_id, state] - 1e8 * (1 - initiation_set)
		best_options = np.flatnonzero(tmp == tmp.max())
		b_probs[best_options] += (1 - epsilon) / len(best_options)
		candidate_option = np.random.choice(num_options, 1, p=b_probs)[0]

		if is_intial_state is True:
			option = candidate_option
			return option, True
		else:
			beta = sigmoid(self.beta_logit[state, prev_option])
			prev_option_term = np.random.choice(2, 1, p=[1 - beta, beta])[0]

			if prev_option_term:
				option = candidate_option
			else:
				option = prev_option
			return option, prev_option_term

	def choose_action(self, state, option):
		"""
		This function samples an action according to the option's policy.
		It returns the action as well as the probability each option chooses that action.
		"""
		probs = softmax(self.pi_logit[state])
		action = np.random.choice(self.num_actions, 1, p=probs[option])[0]
		return action, probs[:, action]

	def choose_option_given_hallway_options_and_actions(self, task_id, state, prev_option, is_intial_state, qs):
		greedy_option = qs[task_id, state].argmax()
		if is_intial_state is True:
			option = greedy_option
			prev_option_term = True
		else:
			if prev_option >= 2:
				beta = 1
			else:
				o1_beta = self.best_option1_beta_vec[state]
				o2_beta = self.best_option2_beta_vec[state]
				tmp = [o1_beta, o2_beta]
				beta = tmp[prev_option]

			prev_option_term = np.random.choice(2, 1, p=[1 - beta, beta])[0]
			if prev_option_term:
				option = greedy_option
			else:
				option = prev_option
		return option, prev_option_term

	def choose_action_given_hallway_options_and_actions(self, state, option):
		if option >= 2:
			action = option - 2
		else:
			o1_action = self.best_option1_pi_vec[state]
			o2_action = self.best_option2_pi_vec[state]
			tmp = [o1_action, o2_action]
			action = tmp[option]
		return action

	def step(self):
		report = {}
		config = self.config
		if self.total_steps < config.phase1_steps:
			storage = Storage(config.rollout_length, ['beta_adv', 'beta_advs', 'adv', 'advs'])

			for _ in range(config.rollout_length):
				task_id = self.task.env.task_id
				epsilon = self.config.epsilon
				option, prev_option_term = self.choose_option(task_id, self.state, self.prev_option, self.s_is_initial_state, self.current_initiation_set, epsilon, self.num_options, self.qs)
				action, option_policy_probs = self.choose_action(self.state, option)
				rhos = option_policy_probs / option_policy_probs[option]

				# env step
				next_state, reward, terminal, info = self.task.step(action)
				next_interests, next_initiation_set = self.choose_initiation_set(next_state)
				ret = info['episodic_return']
				self.compound_return += (reward - config.beta_reg * self.current_interests.sum() * prev_option_term)
				if ret is not None:
					report.setdefault('return', []).append(ret)
					report.setdefault('compound_return', []).append(self.compound_return)
					report.setdefault('episodic_length', []).append(info['episodic_length'])
					self.compound_return = 0

				storage.add({
					'task_id': task_id,
					's': self.state,
					'o': option,
					'a': action,
					'r': reward,
					's_is_initial_state': self.s_is_initial_state,
					'prev_o': self.prev_option,
					'next_s': next_state,
					'next_s_is_initial_state': terminal,
					'rhos': rhos,
					'interests': self.current_interests,
					'initiation_set': self.current_initiation_set,
					'next_initiation_set': next_initiation_set,
					'next_interests': next_interests
				})

				self.prev_option = option
				self.state = next_state
				self.s_is_initial_state = terminal
				self.current_interests = next_interests
				self.current_initiation_set = next_initiation_set
				self.total_steps += 1

			task_id = np.array(storage.task_id)
			s = np.array(storage.s)
			o = np.array(storage.o)
			a = np.array(storage.a)
			rhos = np.array(storage.rhos)
			if config.clip_is is not None:
				rhos = np.clip(rhos, 0, config.clip_is)
			r = np.array(storage.r)
			next_s = np.array(storage.next_s)
			next_s_is_initial_state = np.array(storage.next_s_is_initial_state)
			next_initiation_set = np.array(storage.next_initiation_set)
			next_interests = np.array(storage.next_interests)

			# compute bootstrapping term
			betas = sigmoid(self.beta_logit[next_s[-1]])
			next_v = np.zeros(config.rollout_length)
			next_interests_adv = np.zeros((config.rollout_length, self.num_options))
			for i in range(config.rollout_length):
				next_v[i], next_interests_adv[i] = self.compute_v_bar_and_interests_adv(task_id[i], next_s[i], next_initiation_set[i], next_interests[i])

			rand_number = np.random.rand(self.num_options)
			terminations = rand_number < betas
			rets = (1 - terminations) * self.qs[task_id[-1], next_s[-1]] + terminations * next_v[-1]

			# compute advantages for updating q, pi, and beta
			storage.placeholder()
			for i in reversed(range(config.rollout_length)):
				rets = r[i] + (1 - next_s_is_initial_state[i]) * rets
				advs = rets - self.qs[task_id[i], s[i]]
				storage.advs[i] = advs
				storage.beta_advs[i] = self.qs[task_id[i], next_s[i]] - next_v[i]
			advs = np.array(storage.advs)
			beta_advs = np.array(storage.beta_advs)
			num_adj_options = config.num_adj_options

			for i in range(config.rollout_length):
				# update q
				self.qs[task_id[i], s[i]] += config.q_lr * rhos[i] * advs[i]

				# update interest
				if config.use_interest:

					if next_s_is_initial_state[i] == 1:
						option_term_prob = 1
					else:
						option_term_prob = sigmoid(self.beta_logit[next_s[i], o[i]])
					interest_logit_update = (next_interests[i] * (1 - next_interests[i]) * next_interests_adv[i])[:num_adj_options]
					if config.entropy_reg != 0:
						next_adj_interests = next_interests[i][:num_adj_options]
						interest_logit_update += config.entropy_reg * next_adj_interests * (1 - next_adj_interests) * np.log2((1 - next_adj_interests) / next_adj_interests)  # add entropy
						mask = np.logical_and(next_adj_interests > 1e-8, next_adj_interests < 1 - 1e-8)
						self.interest_logit[next_s[i], :num_adj_options] += config.interest_lr * mask * option_term_prob * interest_logit_update
					else:
						self.interest_logit[next_s[i], :num_adj_options] += config.interest_lr * option_term_prob * interest_logit_update

				# update pi and beta
				# if the option being chosen by the behavior policy is adjustable, update it
				if o[i] < num_adj_options:
					pi_dist = softmax(self.pi_logit[s[i], o[i]])
					pi_update = np.zeros_like(self.pi_logit[s[i], o[i]])
					pi_update[a[i]] += advs[i][o[i]]  # update a[i]'s preference
					pi_update += (- pi_dist) * advs[i][o[i]]  # update other actions preferences
					if config.entropy_reg != 0 and np.all(pi_dist > 1e-8):
						log_pi_dist = np.log2(pi_dist)
						entropy = - np.sum(pi_dist * log_pi_dist)
						pi_update += config.entropy_reg * (- pi_dist) * (log_pi_dist + entropy)  # add entropy
					self.pi_logit[s[i], o[i]] += config.pi_lr * rhos[i][o[i]] * pi_update

					sig = sigmoid(self.beta_logit[next_s[i], o[i]])
					beta_logit_update = sig * (1 - sig) * beta_advs[i][o[i]]
					if config.entropy_reg != 0 and sig > 1e-8 and sig < 1 - 1e-8:
						beta_logit_update += (- config.entropy_reg) * sig * (1 - sig) * np.log2((1 - sig) / sig) # add entropy
					self.beta_logit[next_s[i], o[i]] -= config.beta_lr * rhos[i][o[i]] * beta_logit_update * (1 - next_s_is_initial_state[i])

			if self.total_steps == config.phase1_steps and config.save_model is True:
				save_folder = './experiment/%s/outputs/log/' % self.config.exp_name
				np.save(save_folder + str(self.config.seed) + '_q', self.qs)
				np.save(save_folder + str(self.config.seed) + '_pi', self.pi_logit)
				np.save(save_folder + str(self.config.seed) + '_beta', self.beta_logit)
				np.save(save_folder + str(self.config.seed) + '_interest', self.interest_logit)
		elif self.total_steps < config.phase2_steps:
			if self.total_steps == config.phase1_steps and config.load_model is True:
				load_folder = './experiment/%s/outputs/log/' % self.config.exp_name
				self.qs = np.load(load_folder + str(self.config.seed) + '_q.npy')
				self.pi_logit = np.load(load_folder + str(self.config.seed) + '_pi.npy')
				self.beta_logit = np.load(load_folder + str(self.config.seed) + '_beta.npy')
				self.interest_logit = np.load(load_folder + str(self.config.seed) + '_interest.npy')
			if self.total_steps % 10000==0:
				print("model learning: %d steps" %(self.total_steps))

			self.task.env.env.goals = [-1]
			self.task.env.env.goal_state = -1
			self.task.env.env.task_id = 0
			# model learning
			action = np.random.choice(self.num_actions, 1, p=np.ones(self.num_actions) / self.num_actions)[0]
			# env step
			next_state, reward, terminal, info = self.task.step(action)
			pi = softmax(self.pi_logit[self.state])
			if config.Option_Set_in_VI == 'given_hallway_options_and_actions':
				o1_pi = np.zeros(self.num_actions)
				o2_pi = np.zeros(self.num_actions)
				o1_pi[self.best_option1_pi_vec[self.state]] = 1
				o2_pi[self.best_option2_pi_vec[self.state]] = 1
				pi = np.concatenate(([o1_pi, o2_pi], pi[self.num_adj_options:]))
			elif config.Option_Set_in_VI == 'actions':
				pi = pi[self.num_adj_options:]
			rhos = pi[:, action] / (1 / self.num_actions)
			if terminal and (info['TimeLimit.truncated'] is False):
				raise NotImplementedError
			elif not terminal:
				betas = sigmoid(self.beta_logit[next_state])
				if config.Option_Set_in_VI == 'given_hallway_options_and_actions':
					o1_beta = self.best_option1_beta_vec[next_state]
					o2_beta = self.best_option2_beta_vec[next_state]
					betas = np.concatenate(([o1_beta, o2_beta], betas[self.num_adj_options:]))
				elif config.Option_Set_in_VI == 'actions':
					betas = betas[self.config.num_adj_options:]
				betas_train = np.repeat(np.array([betas]), len(self.train_tasks), axis=0)
				betas_test = np.repeat(np.array([betas]), len(self.test_tasks), axis=0)
				betas_train[np.arange(len(self.train_tasks)), :] = 1 * (np.array([self.train_tasks]).T == next_state) + betas_train[np.arange(len(self.train_tasks)), :] * (np.array([self.train_tasks]).T != next_state)
				betas_test[np.arange(len(self.test_tasks)), :] = 1 * (np.array([self.test_tasks]).T == next_state) + betas_test[np.arange(len(self.test_tasks)), :] * (np.array([self.test_tasks]).T != next_state)
				mat_train = np.zeros((len(self.train_tasks), self.vi_num_options, self.task.env.num_states))
				mat_test = np.zeros((len(self.test_tasks), self.vi_num_options, self.task.env.num_states))
				mat_train[:, :, next_state] = betas_train
				mat_test[:, :, next_state] = betas_test
				self.option_dynamic_model_train[:, self.state, :, :] += config.model_lr * np.expand_dims(rhos, -1) * (mat_train + (1 - np.expand_dims(betas_train, -1)) * self.option_dynamic_model_train[:, next_state, :, :] - self.option_dynamic_model_train[:, self.state, :, :])
				self.option_dynamic_model_test[:, self.state, :, :] += config.model_lr * np.expand_dims(rhos, -1) * (mat_test + (1 - np.expand_dims(betas_test, -1)) * self.option_dynamic_model_test[:, next_state, :, :] - self.option_dynamic_model_test[:, self.state, :, :])
				self.option_reward_model_train[:, self.state, :] += config.model_lr * rhos * (reward + (1 - betas_train) * self.option_reward_model_train[:, next_state, :] - self.option_reward_model_train[:, self.state, :])
				self.option_reward_model_test[:, self.state, :] += config.model_lr * rhos * (reward + (1 - betas_test) * self.option_reward_model_test[:, next_state, :] - self.option_reward_model_test[:, self.state, :])
				self.option_reward_model_train = np.clip(self.option_reward_model_train, -10000, 10000)
				self.option_reward_model_test = np.clip(self.option_reward_model_test, -10000, 10000)
			self.state = next_state
			self.total_steps += 1
			if self.total_steps == config.phase2_steps and config.save_option_model is True:
				save_folder = './experiment/%s/outputs/log/' % self.config.exp_name
				np.save(save_folder + str(self.config.seed) + '_option_dynamic_model_train', self.option_dynamic_model_train)
				np.save(save_folder + str(self.config.seed) + '_option_dynamic_model_test', self.option_dynamic_model_test)
				np.save(save_folder + str(self.config.seed) + '_option_reward_model_train', self.option_reward_model_train)
				np.save(save_folder + str(self.config.seed) + '_option_reward_model_test', self.option_reward_model_test)
		else:
			if self.total_steps == config.phase2_steps and config.load_option_model is True:
				save_folder = './experiment/%s/outputs/log/' % self.config.exp_name
				self.option_dynamic_model_train = np.load(save_folder + str(self.config.seed) + '_option_dynamic_model_train.npy')
				self.option_dynamic_model_test = np.load(save_folder + str(self.config.seed) + '_option_dynamic_model_test.npy')
				self.option_reward_model_train = np.load(save_folder + str(self.config.seed) + '_option_reward_model_train.npy')
				self.option_reward_model_test = np.load(save_folder + str(self.config.seed) + '_option_reward_model_test.npy')

			# planning
			print("planning")
			save_folder = './experiment/%s/outputs/log/' % self.config.exp_name

			# load optimal values
			optimal_values_train = np.load(save_folder + '0_train_value.npy')
			optimal_values_test = np.load(save_folder + '0_test_value.npy')

			num_runs = 30
			planning_iters = 30
			value_error_list = np.zeros((num_runs, planning_iters))
			average_branching_factor_list = np.zeros((num_runs, planning_iters))
			# sampling initiation sets
			for run in range(num_runs):
				print("run ", run)
				if config.Option_Set_in_VI == "learned_options_and_actions":
					self.interest_logit = np.load(save_folder + str(self.config.seed) + '_interest.npy')
					i_sets = np.ones((self.num_states, self.num_options))
					for s in range(self.num_states):
						_, i_set = self.choose_initiation_set(s)
						i_sets[s] = i_set

				# perform planning and evaluate obtained policies.
				self.new_task_q_train = np.random.randn(len(self.train_tasks), self.task.env.num_states,
				                                        self.vi_num_options) + np.ones(
					(len(self.train_tasks), self.task.env.num_states, self.vi_num_options)) * -100
				self.new_task_q_train[np.arange(len(self.train_tasks)), self.train_tasks] = 0  # terminal value is 0
				for i in range(planning_iters):
					if config.vi_add_term_cost:
						term_cost = -config.vi_term_cost
					else:
						term_cost = 0
					# q value iteration
					if config.Option_Set_in_VI == "given_hallway_options_and_actions" or config.Option_Set_in_VI == "actions":
						self.new_task_q_train = self.option_reward_model_train + term_cost + np.sum(
						self.option_dynamic_model_train * np.expand_dims(np.expand_dims(np.max(self.new_task_q_train, 2), 1), 1), 3)
						average_branching_factor_list[run, i] = self.new_task_q_train.shape[-1]
					elif config.Option_Set_in_VI == "learned_options_and_actions":
						update = self.option_reward_model_train + term_cost
						next_q = self.new_task_q_train * i_sets + -100000 * (1 - i_sets)  # options that can not be initiated have very negative values.
						update += np.sum(self.option_dynamic_model_train * np.expand_dims(np.expand_dims(np.max(next_q, 2), 1), 1), 3)
						self.new_task_q_train = update * i_sets + self.new_task_q_train * (1 - i_sets)  # only update options in the initiation set.
						average_branching_factor_list[run, i] = np.mean(i_sets.sum(-1))
					else:
						raise NotImplementedError
					self.new_task_q_train[np.arange(len(self.train_tasks)), self.train_tasks] = 0  # terminal value is 0
					value_error_list[run, i] = np.mean(np.abs(self.new_task_q_train.max(-1) - optimal_values_train))

			title = str(self.config.seed) + '_train'
			np.save(
				save_folder + title,
				np.array([value_error_list, average_branching_factor_list])
			)

			num_runs = 30
			value_error_list = np.zeros((num_runs, planning_iters))
			average_branching_factor_list = np.zeros((num_runs, planning_iters))
			# sampling initiation sets
			for run in range(num_runs):
				print("run ", run)
				if config.Option_Set_in_VI == "learned_options_and_actions":
					self.interest_logit = np.load(save_folder + str(self.config.seed) + '_interest.npy')
					i_sets = np.ones((self.num_states, self.num_options))
					for s in range(self.num_states):
						_, i_set = self.choose_initiation_set(s)
						i_sets[s] = i_set
				self.new_task_q_test = np.random.randn(len(self.test_tasks), self.task.env.num_states,
				                                       self.vi_num_options) + np.ones(
					(len(self.test_tasks), self.task.env.num_states, self.vi_num_options)) * -100
				self.new_task_q_test[np.arange(len(self.test_tasks)), self.test_tasks] = 0
				for i in range(planning_iters):
					if config.vi_add_term_cost:
						term_cost = -1
					else:
						term_cost = 0

					# q value iteration
					if config.Option_Set_in_VI == "given_hallway_options_and_actions" or config.Option_Set_in_VI == "actions":
						self.new_task_q_test = self.option_reward_model_test + term_cost + np.sum(
							self.option_dynamic_model_test * np.expand_dims(
								np.expand_dims(np.max(self.new_task_q_test, 2), 1), 1), 3)
						average_branching_factor_list[run, i] = self.new_task_q_train.shape[-1]
					elif config.Option_Set_in_VI == "learned_options_and_actions":
						update = self.option_reward_model_test + term_cost
						next_q = self.new_task_q_test * i_sets + -100000 * (
									1 - i_sets)  # options that can not be initiated have very negative values.
						update += np.sum(
							self.option_dynamic_model_test * np.expand_dims(np.expand_dims(np.max(next_q, 2), 1), 1), 3)
						self.new_task_q_test = update * i_sets + self.new_task_q_test * (
									1 - i_sets)  # only update options in the initiation set.
						average_branching_factor_list[run, i] = np.mean(i_sets.sum(-1))
					else:
						raise NotImplementedError
					self.new_task_q_test[np.arange(len(self.test_tasks)), self.test_tasks] = 0
					value_error_list[run, i] = np.mean(np.abs(self.new_task_q_test.max(-1) - optimal_values_test))
			save_folder = './experiment/%s/outputs/log/' % self.config.exp_name
			title = str(self.config.seed) + '_test'
			np.save(
				save_folder + title,
				np.array([value_error_list, average_branching_factor_list])
			)
			exit(0)
		return report

	def compute_v_bar_and_interests_adv(self, task_id, s, initiation_set, interests):
		if self.config.use_interest is False:
			epsilon = self.config.epsilon
			return (1 - epsilon) * self.qs[task_id, s].max() + epsilon * self.qs[task_id, s].mean() - np.sum(
				self.config.beta_reg * interests), np.zeros(self.num_options)
		num_adj_options = self.config.num_adj_options
		pos_neg_sets = np.repeat(np.array([initiation_set]), 2 * num_adj_options, axis=0)
		pos_neg_sets[np.arange(num_adj_options), np.arange(num_adj_options)] = 1
		pos_neg_sets[np.arange(num_adj_options) + num_adj_options, np.arange(num_adj_options)] = 0
		epsilon = self.config.epsilon
		k_b_probs = self.compute_k_b_probs(pos_neg_sets, task_id, s, epsilon)
		tmp = np.sum(k_b_probs * self.qs[task_id, s], axis=1)
		v = np.mean(interests[:num_adj_options] * tmp[:num_adj_options] + (1 - interests[:num_adj_options]) * tmp[num_adj_options:]) - np.sum(self.config.beta_reg * interests)

		prob_diff = k_b_probs[:num_adj_options] - k_b_probs[num_adj_options:]
		adv = np.sum(prob_diff * self.qs[task_id, s], axis=1) - self.config.beta_reg
		adv = np.concatenate((adv, np.zeros(self.num_actions)))
		return v, adv

	def eval_step(self, state, reward, terminal, info, ep):
		config = self.config
		interests, initiation_set = self.choose_initiation_set(state)
		greedy_option = (self.qs[self.eval_task.env.task_id, state] - 1e8 * (1 - initiation_set)).argmax()
		if terminal is True:
			option_test = greedy_option
			if ep == 0 and self.total_steps % config.plot_interval == 0:
				if "FourRooms" in config.task_name:
					self.task.env.draw_discovered_options_four_rooms(
						config.exp_name,
						config.sweep_id,
						self.total_steps,
						self.num_options,
						self.get_q,
						self.get_mu,
						self.get_pi,
						self.get_beta,
						self.get_adv_and_interest_adv,
						self.get_interest,
						self.get_mu_over_all_tasks
					)
			option_terminate = -1
		else:
			beta = sigmoid(self.beta_logit[state, self.prev_option_test])
			prev_option_term = np.random.choice(2, 1, p=[1-beta, beta])[0]

			if prev_option_term:
				option_test = greedy_option
				option_terminate = -1
			else:
				option_test = self.prev_option_test
				option_terminate = 0
		action = np.random.choice(self.num_actions, 1, p=softmax(self.pi_logit[state, option_test]))[0]
		self.prev_option_test = option_test
		self.eps_option_list.append(self.prev_option_test)

		interests = sigmoid(self.interest_logit[state])

		return action, option_terminate, option_terminate * np.sum(interests)

	def get_mu(self):
		q = self.qs[self.task.env.task_id]
		mu = np.zeros_like(q)
		mu[np.arange(self.num_states), np.argmax(q, axis=1)] = 1
		return mu

	def get_mu_over_all_tasks(self):
		mu = np.zeros_like(self.qs)
		for task in range(len(self.train_tasks)):
			q = self.qs[task]
			mu[task, np.arange(self.num_states), np.argmax(q, axis=1)] = 1
		mu = np.sum(mu, axis=0) / len(self.train_tasks)
		return mu

	def get_adv_and_interest_adv(self):
		v = np.zeros(self.num_states)
		interest_adv = np.zeros((self.num_states, self.num_options))
		for s in range(self.num_states):
			initiation_set, interests = self.choose_initiation_set(s)
			v[s], interest_adv[s, :] = self.compute_v_bar_and_interests_adv(self.task.env.task_id, s, initiation_set, interests)
		adv = self.qs[self.task.env.task_id] - np.expand_dims(v, 1)
		return adv, interest_adv

	def get_q(self):
		q = self.qs[self.task.env.task_id]
		return q

	def get_pi(self):
		pi = softmax(self.pi_logit)
		return pi

	def get_beta(self):
		beta = sigmoid(self.beta_logit)
		beta = np.stack((1 - beta, beta), axis=2)
		return beta

	def get_interest(self):
		interest = sigmoid(self.interest_logit)
		interest = np.stack((1 - interest, interest), axis=2)
		return interest