import numpy as np
from ..utils.misc import mkdir
from . import user_envs

def make_env(env_id, env_args, seed):
    env_class = getattr(user_envs, env_id)
    env = env_class(env_args)
    env.seed(seed)
    return env


class Env(object):
    def __init__(self,
                 env_id,
                 env_args,
                 log_dir=None,
                 episode_life=True,
                 seed=np.random.randint(int(1e5)),
                 timeout=None
                 ):
        self.env_id = env_id
        self.timeout = timeout
        if log_dir is not None:
            mkdir(log_dir)
        env = make_env(env_id, env_args, seed)
        self.env = env
        self.num_states = self.env.num_states
        self.num_actions = self.env.num_actions
        self._max_episode_steps = timeout
        self._elapsed_steps = 0
        self.eps_return = 0
        self.eps_length = 0

    def step(self, action):

        obs, reward, done, info = self.env.step(action)

        self._elapsed_steps += 1
        info['TimeLimit.truncated'] = False
        if self._elapsed_steps >= self._max_episode_steps and done is False:
            done = True
            info['TimeLimit.truncated'] = True

        self.eps_length += 1
        self.eps_return += reward
        
        if 'TimeLimit.truncated' not in info:
            info['TimeLimit.truncated'] = False
            
        if done:
            info['episodic_return'] = self.eps_return
            info['episodic_length'] = self.eps_length
            self.eps_length = 0
            obs = self.reset()
        else:
            info['episodic_return'] = None
            info['episodic_length'] = self.eps_length
        return obs, reward, done, info

    def reset(self):
        self.eps_return = 0
        self._elapsed_steps = 0
        return self.env.reset()

    def seed(self, seed):
        self.env.seed(seed)