import numpy as np
import time
from pathlib import Path
from ..utils.logger import get_logger


def run_steps(config):
    config.seed = config.sweep_id
    agent = config.agent_class(config)
    agent.logger = get_logger(config.exp_name, config.sweep_id, log_level=config.log_level)
    report_string = "\n"
    for k in config.param_sweeper_dict:
        report_string += "%s: %s \n" % (str(k), str(config.param_sweeper_dict[k]))
    agent.logger.info(report_string)
    total_episodes = 0
    reports = {}
    t0 = time.time()
    last_total_steps = 0
    while True:
        report = agent.step()
        if report is not None:
            for key in report:
                reports.setdefault(key, []).extend(report[key])
        log_flag = int(agent.total_steps / config.log_interval) > int(last_total_steps / config.log_interval)
        eval_flag = int(agent.total_steps / config.eval_interval) > int(last_total_steps / config.eval_interval)
        if log_flag and len(reports) != 0:
            report_string = '\ntotal steps %d\n' % (agent.total_steps)
            for report_name in reports:
                report = reports[report_name]
                if report_name == 'episodic_return':
                    total_episodes += len(reports[report_name])
                    report_string += 'total episodes %3d\n' % (total_episodes)
                if report_name == 'rewards':
                    report_string += 'average reward %3f\n' % (np.sum(report) / config.log_interval)
                if len(report) != 0:
                    report_string += report_name + ' %.3f/%.3f/%.3f/%.3f/%d (mean/median/min/max/num)\n' % (
                        np.mean(report), np.median(report), np.min(report),
                        np.max(report), len(report)
                    )
            report_string += '%.3f steps/s\n' % (config.log_interval / (time.time() - t0))
            agent.logger.info(report_string)
            reports = {}
        t0 = time.time()
        if config.if_eval_episodes and eval_flag:
            episodic_returns, episodic_lengths, episodic_discounted_returns, episodic_discounted_compound_returns, episodic_discounted_terminations, episodic_discounted_interests = agent.eval_episodes()
            report_string = '\nepisodic_return_test' + ' %.3f/%.3f/%.3f/%.3f/%d (mean/median/min/max/num)\n' % (
                np.mean(episodic_returns), np.median(episodic_returns),
                np.min(episodic_returns), np.max(episodic_returns), len(episodic_returns)
            )
            report_string += 'episodic_length_test' + ' %.3f/%.3f/%.3f/%.3f/%d (mean/median/min/max/num)\n' % (
                np.mean(episodic_lengths), np.median(episodic_lengths),
                np.min(episodic_lengths), np.max(episodic_lengths), len(episodic_lengths)
            )
            report_string += 'episodic_discounted_return_test' + ' %.3f/%.3f/%.3f/%.3f/%d (mean/median/min/max/num)\n' % (
                np.mean(episodic_discounted_returns), np.median(episodic_discounted_returns),
                np.min(episodic_discounted_returns), np.max(episodic_discounted_returns), len(episodic_discounted_returns)
            )
            report_string += 'episodic_discounted_compound_returns' + ' %.3f/%.3f/%.3f/%.3f/%d (mean/median/min/max/num)\n' % (
                np.mean(episodic_discounted_compound_returns), np.median(episodic_discounted_compound_returns),
                np.min(episodic_discounted_compound_returns), np.max(episodic_discounted_compound_returns),
                len(episodic_discounted_compound_returns)
            )
            report_string += 'episodic_discounted_terminations_test' + ' %.3f/%.3f/%.3f/%.3f/%d (mean/median/min/max/num)\n' % (
                np.mean(episodic_discounted_terminations), np.median(episodic_discounted_terminations),
                np.min(episodic_discounted_terminations), np.max(episodic_discounted_terminations),
                len(episodic_discounted_terminations)
            )
            report_string += 'episodic_discounted_interests_test' + ' %.3f/%.3f/%.3f/%.3f/%d (mean/median/min/max/num)\n' % (
                np.mean(episodic_discounted_interests), np.median(episodic_discounted_interests),
                np.min(episodic_discounted_interests), np.max(episodic_discounted_interests),
                len(episodic_discounted_interests)
            )
            agent.logger.info(report_string)
            
        if agent.total_steps >= config.max_steps:
            agent.close()
            break
        
        last_total_steps = agent.total_steps

def mkdir(path):
    Path(path).mkdir(parents=True, exist_ok=True)


def close_obj(obj):
    if hasattr(obj, 'close'):
        obj.close()