from algorithm.component import Env
from algorithm import agent
from algorithm.utils.misc import run_steps
from algorithm.utils.param_config import ParamConfig
import os
import numpy as np
import argparse
from alphaex.sweeper import Sweeper

os.environ["SDL_VIDEODRIVER"] = "dummy"


def set_task_fn(cfg):
    cfg.task_fn = lambda seed: Env(cfg.task_name, cfg.task_args, seed=seed, timeout=cfg.timeout)
    cfg.eval_task = cfg.task_fn(seed=0)
    cfg.num_tasks = cfg.eval_task.env.num_tasks


def set_agent_class(cfg):
    cfg.agent_class = getattr(agent, cfg.agent_name)


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="run_file")
    parser.add_argument('--id', default=0, type=int, help='identifies run number and configuration')
    parser.add_argument('--config-file')
    args = parser.parse_args()
    exp_name = args.config_file.split('/')[1]
    project_root = os.path.abspath(os.path.dirname(__file__))
    param_sweeper = Sweeper(os.path.join(project_root, args.config_file))
    param_sweeper_dict = param_sweeper.parse(args.id)
    cfg = ParamConfig(param_sweeper_dict)
    cfg.param_sweeper_dict = param_sweeper_dict
    cfg.exp_name = exp_name
    cfg.data_root = os.path.join(project_root, 'data', 'output')
    cfg.sweep_id = args.id
    np.random.seed(cfg.sweep_id)
    set_task_fn(cfg)
    set_agent_class(cfg)
    bash_script = "mkdir -p experiment/%s/outputs" % cfg.exp_name
    print(bash_script)
    myCmd = os.popen(bash_script).read()
    run_steps(cfg)
